/*  aKode: ALSA Sink

    Copyright (C) 2004 Allan Sandfeld Jensen <kde@carewolf.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <stdint.h>
#include <alsa/asoundlib.h>
#include <alsa/pcm.h>

#include <audioframe.h>
#include "alsa_sink.h"

namespace aKode {

extern "C" { ALSASinkPlugin alsa_sink; }

struct ALSASink::private_data
{
    private_data() : pcm_playback(0), open(false), error(false) {};
    
    snd_pcm_t *pcm_playback;

    AudioConfiguration config;
    int scale;
    uint32_t filled, fragmentSize;
    int sampleSize;
    void* buffer;
    bool open;
    bool error;
};

ALSASink::ALSASink()
{
    m_data = new private_data;
}

ALSASink::~ALSASink()
{
    close();
    delete m_data;
}

bool ALSASink::open()
{
    int err = 0;
    err = snd_pcm_open(&m_data->pcm_playback, "default", SND_PCM_STREAM_PLAYBACK, 0);
    if (err < 0) {
        m_data->error = true;
        return false;
    }

    snd_pcm_nonblock(m_data->pcm_playback, 0);
    
    m_data->error = false;
    return true;
}

void ALSASink::close()
{
    if (m_data->pcm_playback) {
        snd_pcm_drop(m_data->pcm_playback);
        snd_pcm_close(m_data->pcm_playback);
    }
    m_data->pcm_playback = 0;
    m_data->open = false;
    m_data->error = false;
}
    
int ALSASink::setAudioConfiguration(const AudioConfiguration* config)
{
    if (m_data->error) return -1;
    
    int res = 0;
    m_data->config = *config;
    snd_pcm_hw_params_t *hw;
    snd_pcm_hw_params_alloca(&hw);
    snd_pcm_hw_params_any(m_data->pcm_playback, hw);
    snd_pcm_hw_params_set_access(m_data->pcm_playback, hw, SND_PCM_ACCESS_RW_INTERLEAVED);
    //snd_pcm_hw_params_test_format( SND_PCM_FORMAT_S16_LE);
    if (config->sample_width<0) {
        snd_pcm_hw_params_set_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_FLOAT);
        m_data->scale = 1;
    }
    else
    if (config->sample_width<=8) {
        snd_pcm_hw_params_set_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S8);
        m_data->scale = 1<<(8-config->sample_width);
    }
    else
    if (config->sample_width<=16) {
        snd_pcm_hw_params_set_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S16);
        m_data->scale = 1<<(16-config->sample_width);
    }
    else
    if (config->sample_width <=24) {
        snd_pcm_hw_params_set_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S24);
        m_data->scale = 1<<(24-config->sample_width);
    }
    else
    if (config->sample_width <=32) {
        if (snd_pcm_hw_params_test_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S32) == 0) {
            snd_pcm_hw_params_set_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S32);
            m_data->scale = 1<<(32-config->sample_width);
        }
        else
        if (snd_pcm_hw_params_test_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S24) == 0) {
            snd_pcm_hw_params_set_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S24);
            m_data->config.sample_width = 24;
            m_data->scale = 1;
        }
        else
        if (snd_pcm_hw_params_test_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S16) == 0) {
            snd_pcm_hw_params_set_format(m_data->pcm_playback, hw, SND_PCM_FORMAT_S16);
            m_data->config.sample_width = 16;
            m_data->scale = 1;
        }
        else
            res = -1;
    }
    else
        res = -1;

    if (res < 0) {
        m_data->error = true;
        return res;
    }
    
    unsigned int rate = config->sample_rate;
    snd_pcm_hw_params_set_rate_near(m_data->pcm_playback, hw, &rate, 0);
    m_data->config.sample_rate = rate;

    snd_pcm_hw_params_set_channels(m_data->pcm_playback, hw, config->channels);

    //snd_pcm_hw_params_get_period_size(hw, (snd_pcm_uframes_t*)&m_data->fragmentSize, 0);
    m_data->fragmentSize = 1024;
    m_data->buffer = new char[m_data->fragmentSize];
    m_data->filled = 0;
    m_data->open = false;

    if (snd_pcm_hw_params(m_data->pcm_playback, hw) < 0) {
        m_data->error = true;
        return -1;
    }
    else
        return 1;
}

const AudioConfiguration* ALSASink::audioConfiguration() const
{
    return &m_data->config;
}

template<class T>
void ALSASink::_writeFrame(AudioFrame* frame)
{
    int channels = m_data->config.channels;
    int frameSize = sizeof(T)*channels;

    long i = 0;
    T* buffer = (T*)m_data->buffer;
    T** data = (T**)frame->data;
    while(true) {
        if (m_data->filled >= m_data->fragmentSize) {
          xrun:
            int status = snd_pcm_writei(m_data->pcm_playback, m_data->buffer, m_data->fragmentSize/frameSize);
            if (status == -EPIPE) {
                snd_pcm_prepare(m_data->pcm_playback);
                snd_pcm_start(m_data->pcm_playback);
                goto xrun;
            }
            m_data->filled = 0;
        }
        if (i >= frame->length) break;
        for(int j=0; j<channels; j++) {
            buffer[m_data->filled/sizeof(T)] = (data[j][i])*m_data->scale;
            m_data->filled+=sizeof(T);
        }
        i++;
    }
}

bool ALSASink::writeFrame(AudioFrame* frame)
{
    if (m_data->error) return false;
    
    if ( frame->sample_width != m_data->config.sample_width
      || frame->channels != m_data->config.channels
      || frame->sample_rate != m_data->config.sample_rate)
    {
        if (setAudioConfiguration(frame) < 0)
            return false;
    }

    if (!m_data->open) {
        snd_pcm_prepare(m_data->pcm_playback);
        snd_pcm_start(m_data->pcm_playback);
        m_data->open = true;
    }
    
    if (frame->sample_width<0)
        _writeFrame<float>(frame);
    else
    if (frame->sample_width<=8)
        _writeFrame<int8_t>(frame);
    else
    if (frame->sample_width<=16)
        _writeFrame<int16_t>(frame);
    else
    if (frame->sample_width<=32)
        _writeFrame<int32_t>(frame);
    else
        return false;

    return true;
}

} // namespace
